"""
Dual averaging.

This module uses the error codes 80-89.
"""

from typing import Protocol

import jax.numpy as jnp


class DAKernelState(Protocol):
    """
    A protocol for a kernel state with dual averaging support. For an introduction
    to dual averaging, see the blog post by Colin Carroll [#carroll]_ and the Stan
    Reference Manual [#stan]_.

    .. [#carroll] `Colin Carroll, Step Size Adaptation in Hamiltonian Monte Carlo (2019)
       <https://colindcarroll.com/2019/04/21/step-size-adaptation-in-hamiltonian-monte-carlo>`_.
    .. [#stan] `Stan Development Team, Stan Reference Manual (2021), Chapter 15.2
       <https://mc-stan.org/docs/2_28/reference-manual/hmc-algorithm-parameters.html>`_.
    """

    step_size: float
    """The step size of the kernel."""

    error_sum: float
    """The error sum of the acceptance probability. Should not be set by the user,
    but is used by the :func:`.da_step` function."""

    log_avg_step_size: float
    """The logarithm of the average step size. Should not be set by the user, but is
    used by the :func:`.da_step` function."""

    mu: float
    """The bias of the step size proposals. Should not be set by the user, but is
    used by the :func:`.da_step` function."""


def da_init(kernel_state: DAKernelState) -> None:
    """
    Initializes (or resets) a :class:`.DAKernelState`. Returns ``None`` and should be
    called for the side effect on the ``kernel_state`` argument.
    """

    kernel_state.error_sum = 0.0
    kernel_state.log_avg_step_size = jnp.log(kernel_state.step_size)
    kernel_state.mu = jnp.log(10.0 * kernel_state.step_size)


def da_step(
    kernel_state: DAKernelState,
    acceptance_prob: float,
    time_in_epoch: int,
    target_accept: float = 0.8,
    gamma: float = 0.05,
    kappa: float = 0.75,
    t0: int = 10,
) -> None:
    """
    Performs an dual averaging update on a :class:`.DAKernelState`. Returns ``None``
    and should be called for the side effect on the ``kernel_state`` argument.

    Parameters
    ----------
    kernel_state
        A kernel state implementing the :class:`.DAKernelState` protocol.
    acceptance_prob
        The acceptance probability of this MCMC iteration.
    time_in_epoch
        The number of completed MCMC iterations in this epoch.
    target_accept
        The target acceptance probability.
    gamma
        The adaptation regularization scale.
    kappa
        The adaptation relaxation exponent.
    t0
        The adaptation iteration offset.
    """

    ks = kernel_state
    t = time_in_epoch + 1
    eta = t ** (-kappa)

    ks.error_sum += target_accept - acceptance_prob
    log_step_size = ks.mu - (ks.error_sum * jnp.sqrt(t)) / (gamma * (t0 + t))
    ks.step_size = jnp.exp(log_step_size)

    ks.log_avg_step_size = (1 - eta) * ks.log_avg_step_size + eta * log_step_size


def da_finalize(kernel_state: DAKernelState) -> None:
    """
    Sets the new step size in a :class:`.DAKernelState`. Returns ``None`` and should be
    called for the side effect on the ``kernel_state`` argument.
    """

    kernel_state.step_size = jnp.exp(kernel_state.log_avg_step_size)



import jax
import numpy as np
import scipy
from pytest import approx

from liesel.goose.iwls_utils import mvn_log_prob, mvn_sample, solve

rng = np.random.default_rng(1337)


def test_solve():
    n = 5

    lhs = rng.uniform(size=(n, n))
    lhs = 0.5 * (lhs + lhs.T) + n * np.identity(n)
    rhs = rng.uniform(size=n)

    goose_result = solve(np.linalg.cholesky(lhs), rhs)
    np_result = np.linalg.solve(lhs, rhs)

    assert goose_result == approx(np_result)


def test_mvn_log_prob():
    n = 5

    cov = rng.uniform(size=(n, n))
    cov = 0.5 * (cov + cov.T) + n * np.identity(n)
    mean = rng.uniform(size=n)

    x = rng.uniform(size=n)
    goose_result = mvn_log_prob(x, mean, np.linalg.cholesky(np.linalg.inv(cov)))
    scipy_result = scipy.stats.multivariate_normal.logpdf(x, mean, cov)

    assert goose_result == approx(scipy_result)


def test_mvn_random():
    n = 5
    m = 100_000
    seed = 42

    cov = rng.uniform(size=(n, n))
    cov = 0.5 * (cov + cov.T) + n * np.identity(n)
    mean = rng.uniform(size=n)

    prng_key = jax.random.split(jax.random.PRNGKey(seed), m)
    chol_inv_cov = np.linalg.cholesky(np.linalg.inv(cov))

    sample_fn = jax.vmap(lambda prng_key: mvn_sample(prng_key, mean, chol_inv_cov))

    sample = sample_fn(prng_key)

    assert np.cov(sample, rowvar=False) == approx(cov, abs=0.05)



from liesel.goose.epoch import EpochType
from liesel.goose.warmup import stan_epochs


def test_stan_epochs():
    epochs = stan_epochs()
    durations = [epoch.duration for epoch in epochs]
    types = [epoch.type for epoch in epochs]
    i = EpochType.INITIAL_VALUES
    f = EpochType.FAST_ADAPTATION
    s = EpochType.SLOW_ADAPTATION
    p = EpochType.POSTERIOR

    assert durations == [1, 75, 25, 50, 100, 200, 500, 50, 1000]
    assert types == [i, f, s, s, s, s, s, f, p]


def test_stan_epochs_thinning():
    epochs = stan_epochs(thinning_posterior=10, thinning_warmup=5)
    thinning = [epoch.thinning for epoch in epochs]
    types = [epoch.type for epoch in epochs]
    i = EpochType.INITIAL_VALUES
    f = EpochType.FAST_ADAPTATION
    s = EpochType.SLOW_ADAPTATION
    p = EpochType.POSTERIOR

    assert thinning == [1, 5, 5, 5, 5, 5, 5, 5, 10]
    assert types == [i, f, s, s, s, s, s, f, p]
